"""
Pytest fixtures for pywbem end2end tests.
"""

from __future__ import absolute_import

import os
import io
import warnings
from contextlib import contextmanager
import json
import subprocess
try:
    from subprocess import DEVNULL
except ImportError:
    # Python 2.7
    from subprocess import PIPE as DEVNULL
import six
import pytest
import yaml

# Note: The wbem_connection fixture uses the es_server fixture, and
# due to the way py.test searches for fixtures, it also need to be imported.
# pylint: disable=unused-import
from pytest_easy_server import es_server  # noqa: F401
# pylint: enable=unused-import

from pywbem import WBEMServer, Error, ConnectionError, AuthError, \
    ToleratedServerIssueWarning  # pylint: disable=redefined-builtin

from .utils import latest_profile_inst, ServerObjectCache, \
    WBEMConnectionAsserted, server_func_asserted
from .assertions import assert_association_a1, assert_association_a2, \
    assert_association_a3, assert_association_a4, assert_association_a5, \
    assert_association_a6

# Profile definition file (in end2end directory).
PROFILES_YAML_FILE = os.path.join('tests', 'profiles', 'profiles.yml')

# Profile definition list.
# The list items are profile definition items, as described in the profile
# definition file.
with io.open(PROFILES_YAML_FILE, 'r', encoding='utf-8') as _fp:
    PROFILE_DEFINITION_LIST = yaml.safe_load(_fp)
del _fp  # pylint: disable=undefined-loop-variable


@contextmanager
def server_from_docker(image_name, image_port, host_port, container_name,
                       verbose=False):
    """
    Context manager that creates a started WBEM server from a Docker image
    of a WBEM server, if it does not already exist.

    Upon entry, the context manager checks if a container with that name
    exists. If so, it assumes it contains the server and starts the
    container if not currently started. If a container with the name does not
    exist, the image is pulled from DockerHub (using a local Docker cache)
    and a container with that name is created and started.

    Upon exit, the context manager leaves the container running.

    For simplicity of code, the context manager can also be used when no Docker
    image should be used by specifying image_name=None.

    Note: This function requires the 'docker' command to be available locally.

    Parameters:

      image_name (string):
        Docker image name on DockerHub, e.g. org/name:1.2.3. If None, the
        context manager does nothing and returns None.

      image_port (int): Port number in image

      host_port (int): Port number on host (local) system

      container_name (string): Docker container name to use locally

      verbose (bool): Print messages.

    Returns:
      URL of the running WBEM server, or None if image_name=None.
    """
    if image_name:

        # Find out whether container exists and is running
        try:
            inspect_out = subprocess.check_output(
                ['docker', 'container', 'inspect', container_name],
                stderr=DEVNULL)
        except EnvironmentError as exc:
            raise RuntimeError(
                "Docker does not seem to be installed: {}".format(exc))
        except subprocess.CalledProcessError as exc:
            if exc.returncode > 1:
                raise RuntimeError(
                    "Command 'docker container inspect ...' failed: {}".
                    format(exc))
            container_exists = False
            container_running = False
        else:
            container_exists = True
            if isinstance(inspect_out, six.binary_type):
                inspect_out = inspect_out.decode('utf-8')
            _inspect_dict = json.loads(inspect_out)
            _status = _inspect_dict[0]['State']['Status']
            container_running = (_status == 'running')

        if not container_exists:
            docker_pull_with_cache(image_name, verbose=verbose)
            if verbose:
                print("Creating Docker container {} with "
                      "image {}".format(container_name, image_name))
            subprocess.check_call(
                ['docker', 'create',
                 '--name', container_name,
                 '--publish', '{}:{}'.format(host_port, image_port),
                 image_name],
                stdout=DEVNULL)
        else:
            if verbose:
                print("Using existing Docker container {}".
                      format(container_name))

        if not container_running:
            if verbose:
                print("Starting Docker container {}".format(container_name))
            subprocess.check_call(
                ['docker', 'start', container_name],
                stdout=DEVNULL)
        else:
            if verbose:
                print("Docker container {} was already running".
                      format(container_name))

        yield 'https://localhost:{}'.format(host_port)

        # Leave the container running on exit.
    else:
        yield None


def docker_pull_with_cache(image_name, verbose=False):
    """
    Pull a Docker image from DockerHub, using a Docker cache directory.

    If the cache directory does not exist, it is created.

    If the docker image does not exist in the cache directory, it is
    pulled from DockerHub and saved as a TAR file in the cache directory.

    If the docker image exists as a TAR file in the cache directory, it is
    loaded from the TAR file.

    In any case, the result is that the local Docker has the image available.

    The path name of the cache directory is taken from the env var
    DOCKER_CACHE_DIR and defaults to (see code) if not set.

    Parameters:

      image_name (string): Docker image name on DockerHub, e.g. org/name:1.2.3

      verbose (bool): Print messages.
    """
    cache_dir = os.getenv('DOCKER_CACHE_DIR') or '~/docker-cache'
    cache_dir = os.path.expanduser(cache_dir)

    if not os.path.exists(cache_dir):
        if verbose:
            print("Creating local Docker image cache directory: {}".
                  format(cache_dir))
        os.makedirs(cache_dir, 0o755)

    image_tar_fn = image_name.replace('/', '_').replace(':', '_') + '.tar'
    image_tar_path = os.path.join(cache_dir, image_tar_fn)

    if not os.path.exists(image_tar_path):

        if verbose:
            print("Pulling image from Docker Hub: {}".format(image_name))
        subprocess.check_call(
            ['docker', 'pull', image_name],
            stdout=DEVNULL)

        if verbose:
            print("Saving image {} in local Docker image cache: {}".
                  format(image_name, image_tar_path))
        subprocess.check_call(
            ['docker', 'save', '-o', image_tar_path, image_name],
            stdout=DEVNULL)

    else:

        if verbose:
            print("Loading image {} from local Docker image cache: {}".
                  format(image_name, image_tar_path))
        subprocess.check_call(
            ['docker', 'load', '-i', image_tar_path],
            stdout=DEVNULL)


def skip_if_unsupported_capability(conn, cap_name):
    """
    Skip the pytest testcase if the specified capability is not supported
    by the server at the connection.
    """
    if not supports_capability(conn, cap_name):
        pytest.skip("Server {} does not support capability: {}".
                    format(conn.es_server.nickname, cap_name))


def supports_capability(conn, cap_name):
    """
    Return boolean indicating whether the specified capability is supported
    by the server at the connection.
    """
    supported_caps = conn.es_server.user_defined.get('capabilities', [])
    return cap_name in supported_caps


@pytest.fixture(
    scope='function'
)
def wbem_connection(request, es_server):
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Fixture representing a WBEMConnection object to use for the end2end tests.

    The fixture has function scope, so the connection is created for each
    test function (including their parametrizations).

    The WBEM server defined by the 'es_server' parameter is assumed to exist all
    the time during the test session (= the pytest command). If the server
    has a docker image definition, it is established and torn down
    automatically, but only once during the test session.

    Parameters:
      es_server (easy_server.Server): Pytest fixture resolving to the server to
        be tested against.

    Returns:
      pywbem.WBEMConnection: Connection to the server.
    """

    # Display messages about server that is used
    verbose = False

    # Skip this server if we had a skip reason in an earlier attempt.
    # Note: We add an attribute to the es_server object to store this info.
    skip_msg = getattr(es_server, 'skip_msg', None)
    if skip_msg:
        pytest.skip("Remembered skip reason from earlier attempt: {0}".
                    format(skip_msg))

    nickname = es_server.nickname

    # Keep the usage of the items in the es_server object in sync with the
    # schemas defining them in the es_schema.yml file.

    url = es_server.user_defined['url']
    default_namespace = es_server.user_defined.get('default_namespace', None)
    user = es_server.secrets.get('user', None)
    password = es_server.secrets.get('password', None)
    no_verification = es_server.secrets.get('no_verification', None)
    ca_certs = es_server.secrets.get('ca_certs', None)
    cert_file = es_server.secrets.get('cert_file', None)
    key_file = es_server.secrets.get('key_file', None)
    x509 = dict(cert_file=cert_file, key_file=key_file) \
        if cert_file and key_file else None

    image_port = None
    host_port = None
    container_name = None
    image_name = es_server.user_defined.get('docker_image', None)
    if image_name:
        port_mapping = es_server.user_defined.get('docker_port_mapping', None)
        if port_mapping:
            image_port = int(port_mapping['image'])
            host_port = int(port_mapping['host'])
        container_name = 'pywbem_end2endtest_{}'.format(nickname)

    with server_from_docker(
            image_name, image_port, host_port, container_name,
            verbose=verbose):

        if verbose:
            print("Creating WBEM connection to: {}".format(url))

        conn = WBEMConnectionAsserted(
            url, (user, password), x509=x509,
            ca_certs=ca_certs,
            no_verification=no_verification,
            default_namespace=default_namespace,
            timeout=10,
            es_server=es_server)

        # Check that the server can be reached and authenticated with, by
        # issuing some quick operation. Operation failures are tolerated
        # (e.g. GetQualifier is not supported on SFCB), and the connection and
        # authgentication errors can still be detected because they will be
        # detected before the operation fails.
        try:
            conn.GetQualifier(
                'Association',
                asserted=False)
        except ConnectionError as exc:
            msg = "Server {0} at {1}: Server cannot be reached: {2} - {3}". \
                format(nickname, url, exc.__class__.__name__, exc)
            es_server.skip_msg = msg
            warnings.warn(msg, ToleratedServerIssueWarning)
            pytest.skip(msg)
        except AuthError as exc:
            msg = "Server {0} at {1}: Server cannot be authenticated with: " \
                "{2} - {3}". \
                format(nickname, url, exc.__class__.__name__, exc)
            es_server.skip_msg = msg
            warnings.warn(msg, ToleratedServerIssueWarning)
            pytest.skip(msg)
        except Error:
            # Any other error may be due to the WBEM server not supporting
            # qualifier operations. In that case, we don't optimize.
            es_server.skip_msg = None
        else:
            es_server.skip_msg = None

        yield conn

        if verbose:
            print("Closing WBEM connection to: {}".format(url))
        conn.close()


def fixtureid_default_namespace(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `default_namespace()`.

    Parameters:
      * fixture_value (string): The default namespace for the test.
    """
    ns = fixture_value
    return "default_namespace={0}".format(ns)


@pytest.fixture(
    params=[None, 'root/cimv2', 'interop', 'root/interop'],
    scope='module',
    ids=fixtureid_default_namespace
)
def default_namespace(request):
    """
    Fixture representing the set of default namespaces to open connections
    with.

    Returns the default namespace as a string.
    """
    return request.param


def fixtureid_profile_definition(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `profile_definition()`.

    Parameters:
      * fixture_value (dict): The profile definition dictionary of the profile.
    """
    pd = fixture_value
    assert isinstance(pd, dict)
    assert 'registered_name' in pd
    assert 'registered_org' in pd
    return "profile_definition={0}:{1}:{2}". \
        format(pd['registered_org'], pd['registered_name'],
               pd.get('registered_version', 'any'))


def _apply_profile_definition_defaults(pd):
    """Set the profile definition defaults"""
    if 'reference_direction' not in pd:
        pd['reference_direction'] = None
    if 'registered_version' not in pd:
        pd['registered_version'] = None
    if pd['reference_direction'] is None:
        org = pd['registered_org']
        if org == 'SNIA':
            pd['reference_direction'] = 'snia'
        elif org == 'DMTF':
            pd['reference_direction'] = 'dmtf'
        else:
            raise ValueError(
                "Profile definition error: The reference_direction item of a "
                "profile definition can only be defaulted when the registered "
                "organisation is DMTF or SNIA, but it is {0}".format(org))
    if 'scoping_class' not in pd:
        pd['scoping_class'] = None
    if 'scoping_path' not in pd:
        pd['scoping_path'] = None
    return pd


def single_profile_definition(org, name, version=None):
    """
    Return a single profile definition dictionary for the specified org and
    name, and optionally version. If version is None, it matches the
    profile definition with the latest version. Profile definitions that do
    not specify a version always match.
    Optional items in the profile definition dictionary are generated with
    their default values.

    If a profile definition does not exist, returns None.
    """
    assert org is not None
    assert name is not None

    # Determine profile definitions with the specified org and name
    pd_list = []
    for pd in PROFILE_DEFINITION_LIST:
        if org != pd['registered_org'] or name != pd['registered_name']:
            continue
        pd_list.append(pd)

    # Determine profile definition with the specified version or latest
    if version:
        # Search profile definition with that version or without a version
        version_info = version.split('.')
        for pd in pd_list:
            pd_version = pd.get('registered_version', None)
            if not pd_version:
                # A profile definition without a version always matches
                return _apply_profile_definition_defaults(pd.copy())
            pd_version_info = pd_version.split('.')
            if pd_version_info == version_info:
                return _apply_profile_definition_defaults(pd.copy())
    else:
        # Search profile definition with latest version or without a version
        latest_pd_version_info = (-1, -1, -1)
        latest_pd = None
        for pd in pd_list:
            pd_version = pd.get('registered_version', None)
            if not pd_version:
                # A profile definition without a version always matches
                return _apply_profile_definition_defaults(pd.copy())
            pd_version_info = pd_version.split('.')
            if pd_version_info > latest_pd_version_info:
                latest_pd_version_info = pd_version_info
                latest_pd = pd
        if latest_pd:
            return _apply_profile_definition_defaults(latest_pd.copy())

    # No such profile definition exists
    return None


@pytest.fixture(
    params=PROFILE_DEFINITION_LIST,
    scope='module',
    ids=fixtureid_profile_definition
)
def profile_definition(request):
    """
    Fixture representing the set of profile definitions to use for the
    end2end tests.

    Returns the profile definition dictionary of each profile to test against.
    For dict items, see the description in profiles.yaml.

    Optional items in the dictionary are generated with their default values.
    """
    pd = _apply_profile_definition_defaults(request.param.copy())
    return pd


@pytest.fixture(
    params=[
        # the function for a1 must be first, because its results are stored
        # in the object cache for use by the subsequent functions.
        assert_association_a1,
        assert_association_a2,
        assert_association_a3,
        assert_association_a4,
        assert_association_a5,
        assert_association_a6,
    ],
    scope='module',
)
def assert_association_func(request):
    """
    Fixture representing a function for asserting association traversal.

    Returns the function.

    Interface of the function:

        Parameters:
          conn, profile_id, source_path, source_role, assoc_class, far_role,
          far_class

        Returns:
            tuple of:
            - list of associated (far end) instances
            - list of association instances
    """
    return request.param


class ProfileTest(object):
    """
    Base class for end2end tests on a specific profile.
    """

    object_cache = ServerObjectCache()

    def init_profile(self, conn, profile_org, profile_name,
                     profile_version=None):
        """
        Initialize attributes for the profile.

        The following instance attributes are set:
          * conn: conn argument (WBEMConnection).
          * profile_org: profile_org argument (registered org string).
          * profile_name: profile_name argument (registered name string).
          * profile_version: profile_version argument (registered version
            string), may be None.
          * server: WBEMServer object created from the connection.
          * profile_definition: Profile definition dictionary for the profile.
          * profile_id: org:name:version string to identify the profile.
            The version defaults to 'any' if profile_version is None.
          * profile_inst: CIMInstance for the CIM_RegisteredProfile object for
            the profile. If it does not exist, the testcase is skipped.
        """
        # pylint: disable=attribute-defined-outside-init
        assert conn is not None
        assert profile_org is not None
        assert profile_name is not None
        self.conn = conn
        self.profile_org = profile_org
        self.profile_name = profile_name
        self.profile_version = profile_version  # May be None

        self.server = WBEMServer(conn)
        self.profile_definition = single_profile_definition(
            self.profile_org, self.profile_name, self.profile_version)
        assert self.profile_definition is not None

        self.profile_id = "{0}:{1}:{2}".format(
            self.profile_org, self.profile_name, self.profile_version or 'any')
        profile_insts_id = self.profile_id + ':profile_insts'

        try:
            profile_insts = self.object_cache.get_list(
                conn.url, profile_insts_id)
        except KeyError:
            profile_insts = server_func_asserted(
                self.server, 'get_selected_profiles',
                registered_org=self.profile_org,
                registered_name=self.profile_name,
                registered_version=self.profile_version)
            self.object_cache.add_list(
                conn.url, profile_insts_id, profile_insts)

        if not profile_insts:
            pytest.skip("{0} {1} profile (version {2}) is not advertised "
                        "on server {3!r}".
                        format(self.profile_org, self.profile_name,
                               self.profile_version or 'any', self.conn.url))

        self.profile_inst = latest_profile_inst(profile_insts)
        # pylint: enable=attribute-defined-outside-init

    def init_central_instances(self):
        """
        Initialize the central_inst_paths attribute with the central instances
        of the profile. init_profile() must have been called before.

        The following instance attributes are set:
          * central_inst_paths: central instances as list of CIMInstanceName.
        """

        # init_profile() must have been called before.
        assert getattr(self, 'profile_org', None)
        assert getattr(self, 'profile_name', None)

        central_paths_id = self.profile_id + ':central_paths'

        try:
            central_paths = self.object_cache.get_list(
                self.conn.url, central_paths_id)
        except KeyError:
            central_paths = server_func_asserted(
                self.server, 'get_central_instances',
                profile_path=self.profile_inst.path,
                central_class=self.profile_definition['central_class'],
                scoping_class=self.profile_definition['scoping_class'],
                scoping_path=self.profile_definition['scoping_path'],
                reference_direction=self.profile_definition[
                    'reference_direction'])
            self.object_cache.add_list(
                self.conn.url, central_paths_id, central_paths)

        # pylint: disable=attribute-defined-outside-init
        self.central_inst_paths = central_paths



#!/usr/bin/env python
#
# (C) Copyright 2004 Hewlett-Packard Development Company, L.P.
# Copyright 2017 IBM Corp.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Setup script for pywbem project.
"""

import sys
import os
import io
import re
import fnmatch
# setuptools needs to be imported before distutils in order to work.
import setuptools
from distutils import log  # pylint: disable=wrong-import-order
# Just in case, we keep Cython usage conditional on package availability
try:
    from Cython.Build import cythonize
    from Cython.Compiler.Errors import CompileError
except ImportError:
    cythonize = None
from wheel.bdist_wheel import bdist_wheel as bdist_wheel_base


def get_version(version_file):
    """
    Execute the specified version file and return the value of the __version__
    global variable that is set in the version file.

    Note: Make sure the version file does not depend on any packages in the
    requirements list of this package (otherwise it cannot be executed in
    a fresh Python environment).
    """
    with io.open(version_file, 'r', encoding='utf-8') as fp:
        version_source = fp.read()
    _globals = {}
    exec(version_source, _globals)  # pylint: disable=exec-used
    return _globals['__version__']


def get_requirements(requirements_file):
    """
    Parse the specified requirements file and return a list of its non-empty,
    non-comment lines. The returned lines are without any trailing newline
    characters.
    """
    with io.open(requirements_file, 'r', encoding='utf-8') as fp:
        lines = fp.readlines()
    reqs = []
    for line in lines:
        line = line.strip(' \n')
        if not line.startswith('#') and not line.startswith('-r ') and \
                line != '':
            reqs.append(line)
    return reqs


def read_file(a_file):
    """
    Read the specified file and return its content as one string.
    """
    with io.open(a_file, 'r', encoding='utf-8') as fp:
        content = fp.read()
    return content


def find_files(dir_path, fn_pattern):
    """
    Find files in a directory and its subdirectories that match a file name
    pattern.

    Returns a list of path names of the found files. The path names begin
    with the specified directory path name, so if that was absolute or relative,
    the returned file path names will be absolute or relative, respectively.
    """
    found_files = []
    for dirpath, _, filenames in os.walk(dir_path):
        for filename in fnmatch.filter(filenames, fn_pattern):
            found_files.append(os.path.join(dirpath, filename))
    return found_files


def cythonizable_source_files():
    """
    Return the list of path names of files to be cythonized (relative to
    current directory).
    """
    setup_dir = os.path.dirname(__file__)
    # os.path.dirname() returns the empty string on Python 2.7 when run from
    # the work directory, and the empty string is rejected by os.path.relpath().
    if os.path.isabs(setup_dir):
        setup_dir = os.path.relpath(setup_dir)
    py_files = []
    py_files += find_files(os.path.join(setup_dir, 'pywbem'), '*.py')
    py_files += find_files(os.path.join(setup_dir, 'pywbem_mock'), '*.py')
    return py_files


class PytestCommand(setuptools.Command):
    """
    Base class for setup.py commands for executing tests for this package
    using pytest.

    Note on the class name: Because distutils.dist._show_help() shows the class
    name for the setup.py command name instead of invoking get_command_name(),
    the classes that get registered as commands must have the command name.
    """

    description = None  # Set by subclass
    my_test_dirs = None  # Set by subclass

    user_options = [
        (
            'pytest-options=',  # '=' indicates it requires an argument
            None,  # no short option
            "additional options for pytest, as one argument"
        ),
    ]

    def initialize_options(self):
        """
        Standard method called by setup to initialize options for the command.
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_opts = None
        self.test_dirs = None
        self.pytest_options = None
        # pylint: enable=attribute-defined-outside-init

    def finalize_options(self):
        """
        Standard method called by setup to finalize options for the command.
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_opts = [
            '--color=yes',
            '-s',
            '-W', 'default',
            '-W', 'ignore::PendingDeprecationWarning',
        ]
        if sys.version_info[0] == 3:
            self.test_opts.extend([
                '-W', 'ignore::ResourceWarning',
            ])
        self.test_dirs = self.my_test_dirs
        # pylint: enable=attribute-defined-outside-init

    def run(self):
        """
        Standard method called by setup to execute the command.
        """

        # deferred import so install does not depend on it
        import pytest  # pylint: disable=import-outside-toplevel

        args = self.test_opts
        if self.pytest_options:
            args.extend(self.pytest_options.split(' '))
        args.extend(self.test_dirs)

        if self.dry_run:
            self.announce("Dry-run: pytest {}".format(' '.join(args)),
                          level=log.INFO)
            return 0

        self.announce("pytest {}".format(' '.join(args)),
                      level=log.INFO)
        rc = pytest.main(args)
        return rc


class test(PytestCommand):
    # pylint: disable=invalid-name
    """
    Setup.py command for executing unit and function tests.
    """
    description = "pywbem: Run unit and function tests using pytest"
    my_test_dirs = ['tests/unittest', 'tests/functiontest']


class leaktest(PytestCommand):
    # pylint: disable=invalid-name
    """
    Setup.py command for executing leak tests.
    """
    description = "pywbem: Run leak tests using pytest"
    my_test_dirs = ['tests/leaktest']


class end2endtest(PytestCommand):
    # pylint: disable=invalid-name
    """
    Setup.py command for executing end2end tests.
    """
    description = "pywbem: Run end2end tests using pytest"
    my_test_dirs = ['tests/end2endtest']

    def finalize_options(self):
        PytestCommand.finalize_options(self)  # old-style class
        self.test_opts.extend([
            '-v', '--tb=short',
        ])


class bdist_wheel(bdist_wheel_base):
    # pylint: disable=invalid-name
    """
    Replacement for setup.py command 'bdist_wheel' that provides a
    '--cythonized' option that causes the wheel archive to be built as
    a cythonized version.

    The handling of the option is in the module level code because it needs
    to change the setuptools.setup() arguments.
    """
    user_options = bdist_wheel_base.user_options + [
        # (long option, short option, description)
        ('cythonized', None,
         'pywbem: Make a cythonized wheel (default: uncythonized wheel)'),
    ]


# pylint: disable=invalid-name
requirements = get_requirements('requirements.txt')
install_requires = [req for req in requirements
                    if req and not re.match(r'[^:]+://', req)]
dependency_links = [req for req in requirements
                    if req and re.match(r'[^:]+://', req)]
test_requirements = get_requirements('test-requirements.txt')

package_version = get_version(os.path.join('pywbem', '_version.py'))
# pylint: enable=invalid-name

# Docs on setup():
# * https://docs.python.org/2.7/distutils/apiref.html?
#   highlight=setup#distutils.core.setup
# * https://setuptools.readthedocs.io/en/latest/setuptools.html#
#   new-and-changed-setup-keywords
setup_options = dict(
    name='pywbem',
    version=package_version,
    packages=[
        'pywbem',
        'pywbem_mock'
    ],
    include_package_data=True,  # Includes MANIFEST.in files into sdist (only)
    scripts=[
        'mof_compiler', 'mof_compiler.bat'
    ],
    requires=['setuptools', 'wheel'],
    install_requires=install_requires,
    dependency_links=dependency_links,
    extras_require={
        "test": test_requirements,
    },
    cmdclass={
        'test': test,
        'leaktest': leaktest,
        'end2endtest': end2endtest,
        'bdist_wheel': bdist_wheel,
    },
    description='pywbem - A WBEM client',
    long_description=read_file('README_PYPI.rst'),
    long_description_content_type='text/x-rst',
    license='LGPL version 2.1, or (at your option) any later version',
    author='Tim Potter',
    author_email='tpot@hp.com',
    maintainer='Andreas Maier, Karl Schopmeyer',
    maintainer_email='maiera@de.ibm.com, k.schopmeyer@swbell.net',
    url='https://pywbem.github.io/pywbem/',
    project_urls={
        'Bug Tracker': 'https://github.com/pywbem/pywbem/issues',
        'Documentation': 'https://pywbem.readthedocs.io/en/latest/',
        'Source Code': 'https://github.com/pywbem/pywbem',
    },

    options={'bdist_wheel': {'universal': True}},
    zip_safe=True,  # This package can safely be installed from a zip file
    platforms='any',

    # Keep these Python versions in sync with pywbem/__init__.py
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Systems Administration',
    ]
)

CYTHONIZED = False
if cythonize and '--cythonized' in sys.argv:
    sys.argv.remove('--cythonized')
    if 'bdist_wheel' in sys.argv:
        CYTHONIZED = True

if CYTHONIZED:
    src_files = cythonizable_source_files()
    print("Cythonizing pywbem source files")
    if '--dry-run' in sys.argv:
        for fn in src_files:
            print("Dry-run: Cythonizing: {}".format(fn))
        cython_ext_modules = []
    else:
        try:
            cython_ext_modules = cythonize(
                src_files,
                build_dir='build_cythonize',
                annotate=False,
                compiler_directives=dict(
                    language_level=str(sys.version_info[0])))
        except CompileError as exc:
            print("Error: Cython compile error: {}".format(exc))
            sys.exit(1)

    setup_options['requires'] += ['Cython']
    setup_options['ext_modules'] = cython_ext_modules
    setup_options['options']['build'] = {'build_lib': 'build_build_ext'}
    setup_options['zip_safe'] = False

setuptools.setup(**setup_options)



"""
Utility functions for creating random CIM objects.
"""

from collections import namedtuple
import string
import random
from datetime import datetime
import six

from pywbem import CIMQualifier, CIMProperty, CIMClass, CIMInstance, \
    CIMInstanceName, CIMDateTime
from pywbem_mock import FakedWBEMConnection


ASCII_CHARS = [six.unichr(i) for i in range(0x0020, 0x007e + 1)]

#: The CIM integer type names
INTEGER_TYPES = (
    'uint8', 'uint16', 'uint32', 'uint64',
    'sint8', 'sint16', 'sint32', 'sint64',
)

#: The CIM floating point type names
REAL_TYPES = ('real32', 'real64')


#: Range of integer values, for various purposes
#: Attributes:
#: - min (int): Minimum value (inclusive)
#: - min (int): Maximum value (inclusive)
Range = namedtuple(
    'Range',
    ['min', 'max']
)

#: Profile for generating typed CIM values
#: Attributes:
#: - types (tuple of string): CIM type names
#: - string_len (Range): Lengths of string values
ValueProfile = namedtuple(
    'ValueProfile',
    ['types', 'string_len']
)

#: A meaningful set of CIM qualifier values
#: Attributes:
#: - names (tuple of string): Names of the qualifiers in the set
#: - types (tuple of string: CIM type names for the rs in the set
#: - values (ValueProfile or value): Qualifier values
QualifierSet = namedtuple(
    'QualifierSet',
    ['names', 'types', 'values']
)

#: Profile for picking a random set of CIM qualifier values
#: Attributes:
#: - qualifier_sets (tuple of QualifierSet): Qualifier sets to pick from
QualifierSetProfile = namedtuple(
    'QualifierSetProfile',
    ['qualifier_sets']
)

#: Profile for generating multiple CIM properties for use in CIM classes
#: Attributes:
#: - name_len (Range): Lengths of property names
#: - value_profile (ValueProfile): Property value profile
#: - value_ratio (float): Ratio of properties that have a value
#: - description_len (Range): Lengths of property Description qualifier;
#:   None for no Description qualifier
#: - qualifier_set_profile (QualifierSetProfile): Qualifier set profile;
#:   None for no additional qualifiers
PropertyProfile = namedtuple(
    'PropertyProfile',
    ['name_len', 'value_profile', 'value_ratio', 'description_len',
     'qualifier_set_profile']
)


def random_name(length_range):
    """
    Return a random valid CIM name of a length that is randomly in a length
    range.

    Parameters:
      length_range (Range): Length range.
    """
    first_chars = string.ascii_uppercase
    other_chars = string.ascii_uppercase + string.ascii_lowercase + \
        string.digits
    length = random.randint(length_range.min, length_range.max)
    name = random.choice(first_chars) + ''.join(
        random.choice(other_chars) for _ in range(length))
    return name


def random_string(length_range, charset='ascii'):
    """
    Return a random Unicode string value of a length that is randomly in a
    length range, and the string characters are in a character set.

    Parameters:
      length_range (Range): Length range.
      charset (string): CHaracter set for the values. Valid character sets
        are:
        - 'ascii' - Printable 7-bit ASCII characters (U+0020 .. U+007E)
    """
    if charset == 'ascii':
        chars = ASCII_CHARS
    else:
        raise ValueError("Invalid charset: {}".format(charset))
    length = random.randint(length_range.min, length_range.max)
    value = ''.join(random.choice(chars) for _ in range(length))
    return value


def random_type_value(values, type=None):
    # pylint: disable=redefined-builtin
    """
    Return a random valid tuple of CIM type name and value according to
    a value profile.

    Parameters:
      values (ValueProfile or value): Values
      type (string): Use this type instead of a random type from the profile
    """

    if not isinstance(values, ValueProfile):
        if type is None:
            raise ValueError("Specifying a value directly requires specifying "
                             "a type")
        return type, values

    if type is None:
        type = random.choice(values.types)

    if type == 'string':
        value = random_string(values.string_len)
    elif type == 'char16':
        value = random_string(Range(1, 1))
    elif type in INTEGER_TYPES:
        # TODO: Consider value range of type
        value = random.randint(0, 127)
    elif type in REAL_TYPES:
        # TODO: Consider value range of type
        value = float(random.randint(-1000, 1000))
    elif type in 'boolean':
        value = random.choice((True, False))
    elif type in 'datetime':
        year = random.randint(1, 3000)
        month = random.randint(1, 12)
        day = random.randint(1, 31)
        hour = random.randint(0, 23)
        minute = random.randint(0, 59)
        second = random.randint(0, 59)
        value = CIMDateTime(datetime(year, month, day, hour, minute, second))
    else:
        # The 'reference' type is intentionally not supported here
        raise AssertionError('Invalid CIM type name: {}'.format(type))
    return type, value


def make_properties(number, property_profile):
    """
    Construct and return a list of CIMProperty objects for use on CIM classes,
    that has the specified number of properties according to the specified
    property profile.
    """

    props = []
    for _ in range(0, number):
        pname = random_name(property_profile.name_len)
        ptype, pvalue = random_type_value(property_profile.value_profile)
        if random.random() > property_profile.value_ratio:
            pvalue = None
        pquals = []
        if property_profile.description_len:
            pquals.append(CIMQualifier(
                'Description', type='string',
                value=random_string(property_profile.description_len)))
        if property_profile.qualifier_set_profile:
            qset = random.choice(
                property_profile.qualifier_set_profile.qualifier_sets)
            for j, qname in enumerate(qset.names):
                qtype = qset.types[j]
                _, qvalue = random_type_value(qset.values, type=qtype)
                pquals.append(CIMQualifier(qname, type=qtype, value=qvalue))
        prop = CIMProperty(pname, type=ptype, value=pvalue, qualifiers=pquals)
        props.append(prop)
    return props


def make_class(properties):
    """
    Construct and return a CIMClass object from the specified properties.
    """
    cname = random_name(Range(8, 16))
    cls = CIMClass(cname, properties=properties)
    return cls


def make_instances(cls, number, value_profile):
    """
    Construct and return a number of CIMInstance objects with path, with
    random values for the instance properties.
    """
    insts = []
    for _ in range(0, number):
        inst = CIMInstance(cls.classname)
        # TODO: Make namespace flexible
        inst.path = CIMInstanceName(cls.classname, namespace='root/cimv2')
        for pname, cls_prop in cls.properties.items():
            ptype = cls_prop.type
            _, pvalue = random_type_value(value_profile, type=ptype)
            inst_prop = CIMProperty(pname, type=ptype, value=pvalue)
            inst.properties[pname] = inst_prop
            if cls_prop.qualifiers.get('Key', False):
                inst.path.keybindings[pname] = pvalue
        insts.append(inst)
    return insts


def tst_random_class(num_props):
    """
    Return a CIMClass object with a random class definition that has the
    specified number of properties.
    """
    prop_qset_key = QualifierSet(
        ('Key',),
        ('boolean',),
        True)
    prop_qset_string = QualifierSet(
        ('MinLen', 'MaxLen'),
        ('uint32', 'uint32'),
        ValueProfile(('uint32',), None))
    prop_qset_uint32 = QualifierSet(
        ('MinValue', 'MaxValue'),
        ('sint64', 'sint64'),
        ValueProfile(('sint64',), None))
    prop_profile_key = PropertyProfile(
        Range(4, 16),  # name_len
        ValueProfile(('string',), Range(8, 64)),
        0,  # value_ratio
        Range(16, 256),  # description_len
        QualifierSetProfile((prop_qset_key,))
    )
    prop_profile_string = PropertyProfile(
        Range(4, 16),  # name_len
        ValueProfile(('string',), Range(8, 64)),
        0.2,  # value_ratio
        Range(16, 256),  # description_len
        QualifierSetProfile((prop_qset_string,))
    )
    prop_profile_uint32 = PropertyProfile(
        Range(4, 16),  # name_len
        ValueProfile(('uint32',), None),
        0.2,  # value_ratio
        Range(16, 256),  # description_len
        QualifierSetProfile((prop_qset_uint32,))
    )
    prop_profile_boolean = PropertyProfile(
        Range(4, 16),  # name_len
        ValueProfile(('boolean',), None),
        0.2,  # value_ratio
        Range(16, 256),  # description_len
        None  # no qualifiers
    )
    rem_props = num_props
    num_key = 1
    rem_props -= num_key
    num_string = int(rem_props * 0.4)
    rem_props -= num_string
    num_uint32 = int(rem_props * 0.6)
    rem_props -= num_uint32
    num_boolean = rem_props
    props = []
    props.extend(make_properties(num_key, prop_profile_key))
    props.extend(make_properties(num_string, prop_profile_string))
    props.extend(make_properties(num_uint32, prop_profile_uint32))
    props.extend(make_properties(num_boolean, prop_profile_boolean))
    cls = make_class(props)
    return cls


def tst_random_instances(cls, num_insts):
    """
    Return a list of the specified number of CIMInstance objects of the
    specified class with random property values.
    """
    value_profile_all = ValueProfile(
        ('string', 'uint32', 'boolean'),
        Range(8, 64))
    instances = make_instances(cls, num_insts, value_profile_all)
    return instances


def tst_random_conn(num_insts, num_props):
    """
    Build a FakedWBEMConnection with a mock environment that contains:
    - the qualifiers needed for the class
    - a class with the specified number of properties of types string, uint32,
      boolean, with one string-typed key property
    - the specified number of instances of that class, with random property
      values

    Returns:
      tuple(FakedWBEMConnection, CIMClass, list(CIMInstance))
    """

    qualifiers_mof = """
Qualifier Description : string = null,
    Scope(any),
    Flavor(EnableOverride, ToSubclass, Translatable);

Qualifier Key : boolean = false,
    Scope(property, reference),
    Flavor(DisableOverride, ToSubclass);

Qualifier MaxLen : uint32 = null,
    Scope(property, method, parameter);

Qualifier MaxValue : sint64 = null,
    Scope(property, method, parameter);

Qualifier MinLen : uint32 = 0,
    Scope(property, method, parameter);

Qualifier MinValue : sint64 = null,
    Scope(property, method, parameter);
    """

    cls = tst_random_class(num_props)
    instances = tst_random_instances(cls, num_insts)

    # Build mock environment (using default namespace)
    conn = FakedWBEMConnection(disable_pull_operations=True)
    conn.compile_mof_string(qualifiers_mof)
    conn.add_cimobjects(cls)
    conn.add_cimobjects(instances)

    return conn, cls, instances



#!/usr/bin/env python

"""
Tests for statistics (`_statistics` in pywbem module).
"""

from __future__ import absolute_import, print_function

import re
import time
import pytest

from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import Statistics  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


@pytest.fixture(params=[
    True,
    False,
], scope='module')
def statistics_enable(request):
    """
    Possible values for Statistics(enable).
    """
    return request.param


def time_abs_delta(t1, t2):
    """
    Return the positive difference between two float values.
    """
    return abs(t1 - t2)


TESTCASES_STATISTICS_INIT = [

    # Testcases for Statistics.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to CIMInstanceName().
    #   * init_kwargs: Dict of keyword arguments to CIMInstanceName().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Verify default arguments",
        dict(
            init_args=[],
            init_kwargs={},
            exp_attrs=dict(
                enabled=False,
            ),
        ),
        None, None, True
    ),
    (
        "Verify order of positional arguments",
        dict(
            init_args=[
                True,
            ],
            init_kwargs={},
            exp_attrs=dict(
                enabled=True,
            ),
        ),
        None, None, True
    ),
    (
        "Verify names of arguments",
        dict(
            init_args=[],
            init_kwargs=dict(
                enable=True,
            ),
            exp_attrs=dict(
                enabled=True,
            ),
        ),
        None, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_STATISTICS_INIT)
@simplified_test_function
def test_Statistics_init(testcase, init_args, init_kwargs, exp_attrs):
    """
    Test function for Statistics.__init__()
    """

    # The code to be tested
    statistics = Statistics(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_enabled = exp_attrs['enabled']
    assert statistics.enabled == exp_enabled
    assert isinstance(statistics.enabled, type(exp_enabled))


def test_Statistics_enable(statistics_enable):
    # pylint: disable=redefined-outer-name
    """
    Test function for Statistics.enable()
    """

    statistics = Statistics(enable=statistics_enable)

    # The code to be tested
    statistics.enable()

    assert statistics.enabled is True


def test_Statistics_disable(statistics_enable):
    # pylint: disable=redefined-outer-name
    """
    Test function for Statistics.disable()
    """

    statistics = Statistics(enable=statistics_enable)

    # The code to be tested
    statistics.disable()

    assert statistics.enabled is False


TESTCASES_STATISTICS_GET_OP_STATISTIC = [

    # Testcases for Statistics.get_op_statistic()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to CIMInstanceName().
    #   * init_kwargs: Dict of keyword arguments to CIMInstanceName().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Verify initial stats and snapshot for disabled statistics",
        dict(
            init_enable=False,
            method_name='EnumerateInstances',
            exp_snapshot_len=0,
            exp_op_stats_attrs=dict(
                name='disabled',
                count=0,
                avg_time=0,
                min_time=float('inf'),
                max_time=0,
                avg_request_len=0,
                min_request_len=float('inf'),
                max_request_len=0,
            ),
        ),
        None, None, True
    ),
    (
        "Verify initial stats and snapshot for enabled statistics",
        dict(
            init_enable=True,
            method_name='EnumerateInstances',
            exp_snapshot_len=1,
            exp_op_stats_attrs=dict(
                name='EnumerateInstances',
                count=0,
                avg_time=0,
                min_time=float('inf'),
                max_time=0,
                avg_request_len=0,
                min_request_len=float('inf'),
                max_request_len=0,
            ),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_STATISTICS_GET_OP_STATISTIC)
@simplified_test_function
def test_Statistics_get_op_statistic(
        testcase, init_enable, method_name, exp_snapshot_len,
        exp_op_stats_attrs):
    # pylint: disable=unused-argument
    """
    Test function for Statistics.get_op_statistic()
    """

    statistics = Statistics(enable=init_enable)

    # The code to be tested
    op_stats = statistics.get_op_statistic(method_name)

    snapshot_length = len(statistics.snapshot())
    assert snapshot_length == exp_snapshot_len

    for attr_name in exp_op_stats_attrs:
        exp_attr_value = exp_op_stats_attrs[attr_name]
        attr_value = getattr(op_stats, attr_name)
        assert attr_value == exp_attr_value, \
            "Unexpected op_stats attribute '{}'".format(attr_name)


def test_Statistics_measure_enabled():
    """
    Test measuring time with enabled statistics.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(100, 200)

    for _, stats in statistics.snapshot():
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert stats.max_request_len == 100
        assert stats.min_request_len == 100
        assert stats.avg_request_len == 100
        assert stats.max_reply_len == 200
        assert stats.min_reply_len == 200
        assert stats.avg_reply_len == 200

    stats.reset()
    assert stats.count == 0
    assert stats.avg_time == 0
    assert stats.min_time == float('inf')
    assert stats.max_time == 0


def test_Statistics_measure_disabled_cm():
    """
    Test measuring time with disabled statistics, via context manager.
    """

    statistics = Statistics()

    duration = 0.1

    with statistics('EnumerateInstances'):
        time.sleep(duration)

    stats_list = statistics.snapshot()
    assert len(stats_list) == 0


def test_Statistics_measure_enabled_cm():
    """
    Test measuring time with enabled statistics, via context manager.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 0.1

    with statistics('EnumerateInstances'):
        time.sleep(duration)

    stats_dict = dict(statistics.snapshot())
    assert len(stats_dict) == 1

    assert 'EnumerateInstances' in stats_dict
    stats = stats_dict['EnumerateInstances']
    assert stats.count == 1


def test_Statistics_measure_enabled_nested_cm():
    """
    Test measuring time with enabled statistics, via nested context managers.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 0.1
    inner_count = 3

    with statistics('compile_schema_classes'):
        for _ in range(0, inner_count):
            with statistics('compile_mof_string'):
                time.sleep(duration)

    stats_dict = dict(statistics.snapshot())
    assert len(stats_dict) == 2

    assert 'compile_schema_classes' in stats_dict
    stats = stats_dict['compile_schema_classes']
    assert stats.count == 1

    assert 'compile_mof_string' in stats_dict
    stats = stats_dict['compile_mof_string']
    assert stats.count == inner_count


def test_Statistics_measure_enabled_with_servertime():
    # pylint: disable=invalid-name
    """
    Test measuring time with enabled statistics.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(1000, 2000, duration)

    for _, stats in statistics.snapshot():
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) <= delta
        assert time_abs_delta(stats.min_time, duration) <= delta
        assert time_abs_delta(stats.max_time, duration) <= delta

        assert time_abs_delta(stats.avg_server_time, duration) <= delta
        assert time_abs_delta(stats.min_server_time, duration) <= delta
        assert time_abs_delta(stats.max_server_time, duration) <= delta

        assert stats.max_request_len == 1000
        assert stats.min_request_len == 1000
        assert stats.avg_request_len == 1000
        assert stats.max_reply_len == 2000
        assert stats.min_reply_len == 2000
        assert stats.avg_reply_len == 2000

    stats.reset()
    assert stats.count == 0
    assert stats.avg_time == 0
    assert stats.min_time == float('inf')
    assert stats.max_time == 0


def test_Statistics_measure_disabled():
    """
    Test measuring time with disabled statistics.
    """

    statistics = Statistics()

    duration = 0.2

    stats = statistics.get_op_statistic('GetClass')
    assert stats.name == 'disabled'

    stats.start_timer()
    time.sleep(duration)
    stats.stop_timer(100, 200)

    for _, stats in statistics.snapshot():
        assert stats.count == 0
        assert stats.avg_time == 0
        assert stats.min_time == float('inf')
        assert stats.max_time == 0


def test_Statistics_measure_avg():
    """
    Test measuring time with enabled statistics.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.8

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(100, 200)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(200, 400)

    for _, stats in statistics.snapshot():
        assert stats.count == 2
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert stats.max_request_len == 200
        assert stats.min_request_len == 100
        assert stats.avg_request_len == 150
        assert stats.max_reply_len == 400
        assert stats.min_reply_len == 200
        assert stats.avg_reply_len == 300


def test_Statistics_measure_exception():
    """
    Test measuring time with enabled statistics.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(100, 200)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(duration)
    stats.stop_timer(200, 400)

    for _, stats in statistics.snapshot():
        assert stats.count == 2
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert stats.max_request_len == 200
        assert stats.min_request_len == 100
        assert stats.avg_request_len == 150
        assert stats.max_reply_len == 400
        assert stats.min_reply_len == 200
        assert stats.avg_reply_len == 300


def test_Statistics_snapshot():
    """
    Test that snapshot() takes a stable snapshot.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('GetInstance')
    time.sleep(duration)
    stats.stop_timer(100, 200)

    # take the snapshot
    snapshot = statistics.snapshot()

    # keep producing statistics data
    stats.start_timer()
    time.sleep(duration)
    stats.stop_timer(100, 200)

    # verify that only the first set of data is in the snapshot
    for _, stats in snapshot:
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta


def test_Statistics_reset():
    """
    Test resetting statistics.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('GetInstance')
    # test reset fails because stat in process
    assert statistics.reset() is False
    time.sleep(duration)
    stats.stop_timer(100, 200)

    # take a snapshot
    snapshot = statistics.snapshot()

    # verify that only the first set of data is in the snapshot
    for _, stats in snapshot:
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta

    assert statistics.reset() is True

    # take another snapshot. This snapshot should be empty
    snapshot = statistics.snapshot()
    assert not snapshot


def test_Statistics_server_time_suspension():
    """
    Test suspending server time and resetting.

    Server time suspension occurs at the level of operations (i.e.
    OperationStatistics object) if not all executions of the operation return
    the server response time. Once it has occurred, the server time values
    are reset and this condition is remembered until the OperationStatistics
    object's reset() is called (which normally is not used and not tested here)
    or the parent Statistic object's reset() is called.
    """

    statistics = Statistics()
    statistics.enable()

    duration = 1.0
    server_resp_time = 0.8

    # Allowable delta in seconds between expected and actual duration.
    # Notes:
    # * Windows has only a precision of 1/60 sec.
    # * In CI environments, the tests sometimes run slow.
    delta = 0.5

    stats = statistics.start_timer('GetInstance')
    time.sleep(duration)
    stats.stop_timer(100, 200, server_resp_time)

    # verify server time after server response time was provided
    snapshot = statistics.snapshot()
    for _, stats in snapshot:
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert time_abs_delta(stats.avg_server_time, server_resp_time) < delta
        assert time_abs_delta(stats.min_server_time, server_resp_time) < delta
        assert time_abs_delta(stats.max_server_time, server_resp_time) < delta

    stats = statistics.start_timer('GetInstance')
    time.sleep(duration)
    stats.stop_timer(120, 250, None)

    # verify server time (and only that) is suspended when server response
    # time was not provided
    snapshot = statistics.snapshot()
    for _, stats in snapshot:
        assert stats.count == 2
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert stats.avg_server_time == float(0)
        assert stats.min_server_time == float('inf')
        assert stats.max_server_time == float(0)

    stats = statistics.start_timer('GetInstance')
    time.sleep(duration)
    stats.stop_timer(120, 250, server_resp_time)

    # verify that server time suspension is sticky once it occurred, even if
    # server response time is provided again
    snapshot = statistics.snapshot()
    for _, stats in snapshot:
        assert stats.count == 3
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert stats.avg_server_time == float(0)
        assert stats.min_server_time == float('inf')
        assert stats.max_server_time == float(0)

    assert statistics.reset() is True

    stats = statistics.start_timer('GetInstance')
    time.sleep(duration)
    stats.stop_timer(100, 200, server_resp_time)

    # verify that reset() also resets server time suspension
    snapshot = statistics.snapshot()
    for _, stats in snapshot:
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, duration) < delta
        assert time_abs_delta(stats.min_time, duration) < delta
        assert time_abs_delta(stats.max_time, duration) < delta
        assert time_abs_delta(stats.avg_server_time, server_resp_time) < delta
        assert time_abs_delta(stats.min_server_time, server_resp_time) < delta
        assert time_abs_delta(stats.max_server_time, server_resp_time) < delta


def test_Statistics_print_statistics():
    """
    Test repr() and formatted() for a small statistics.
    """

    statistics = Statistics()
    statistics.enable()

    stats = statistics.start_timer('EnumerateInstanceNames')
    time.sleep(0.1)
    stats.stop_timer(1200, 22000)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.1)
    stats.stop_timer(1000, 20000)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.2)
    stats.stop_timer(1500, 25000)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.4)
    stats.stop_timer(1200, 35000)

    # test repr output
    stat_repr = repr(statistics)

    assert re.match(
        r'Statistics\(',
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstanceNames', count=1,"
        r" exception_count=0, avg_time=[.0-9]+, min_time=[.0-9]+, "
        r"max_time=[.0-9]+, avg_server_time=0.0, min_server_time=inf, "
        r"max_server_time=0.0, avg_request_len=[.0-9]+, "
        r"min_request_len=[0-9]{4}, max_request_len=[0-9]{4}, "
        r"avg_reply_len=[.0-9]+, min_reply_len=[0-9]{5},"
        r" max_reply_len=[0-9]{5}",
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstances', count=3, "
        r"exception_count=0, avg_time=[.0-9]+, min_time=[.0-9]+, "
        r"max_time=[.0-9]+, avg_server_time=0.0, min_server_time=inf, "
        r"max_server_time=0.0, avg_request_len=[.0-9]+, "
        r"min_request_len=[0-9]{4}, max_request_len=[0-9]{4}, "
        r"avg_reply_len=[.0-9]+, min_reply_len=[0-9]{5}, "
        r"max_reply_len=[0-9]{5}",
        stat_repr)

    # Test statistics report output

    report = statistics.formatted()

    assert re.match(
        r'Statistics',
        report)

    assert re.search(
        r"Count ExcCnt +Time \[s\] +RequestLen \[B\] +ReplyLen \[B\] "
        r"+Operation",
        report)

    assert re.search(
        r" +3 +0 +[.0-9]+ +[.0-9]+ +[.0-9]+ +"
        r"[.0-9]+ +[0-9]{4} +[0-9]{4} +"
        r"[.0-9]+ +[0-9]{5} +[0-9]{5} EnumerateInstances",
        report)

    assert re.search(
        r" +1 +0 +[.0-9]+ +[.0-9]+ +[.0-9]+ +"
        r"[.0-9]+ +[0-9]{4} +[0-9]{4} +"
        r"[.0-9]+ +[0-9]{5} +[0-9]{5} EnumerateInstanceNames",
        report)


def test_Statistics_print_stats_svrtime():
    """
    Test repr() and formatted() for a small statistics.
    """

    statistics = Statistics()
    statistics.enable()

    stats = statistics.start_timer('EnumerateInstanceNames')
    time.sleep(0.1)
    stats.stop_timer(1200, 22000, 0.1)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.1)
    stats.stop_timer(1000, 20000, 0.1)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.2)
    stats.stop_timer(1500, 25000, 0.2)

    stats = statistics.start_timer('EnumerateInstances')
    time.sleep(0.4)
    stats.stop_timer(1200, 35000, 0.4)

    # test repr output
    stat_repr = repr(statistics)

    # test repr output
    assert re.match(
        r'Statistics\(',
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstanceNames', count=1,"
        r" exception_count=0, avg_time=[.0-9]+, min_time=[.0-9]+, "
        r"max_time=[.0-9]+, avg_server_time=[.0-9]+, "
        r"min_server_time=[.0-9]+, "
        r"max_server_time=[.0-9]+, avg_request_len=[.0-9]+, "
        r"min_request_len=[0-9]{4}, max_request_len=[0-9]{4}, "
        r"avg_reply_len=[.0-9]+, min_reply_len=[0-9]{5},"
        r" max_reply_len=[0-9]{5}",
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstances', count=3, "
        r"exception_count=0, avg_time=[.0-9]+, min_time=[.0-9]+, "
        r"max_time=[.0-9]+, avg_server_time=[.0-9]+, "
        r"min_server_time=[.0-9]+, "
        r"max_server_time=[.0-9]+, avg_request_len=[.0-9]+, "
        r"min_request_len=[0-9]{4}, max_request_len=[0-9]{4}, "
        r"avg_reply_len=[.0-9]+, min_reply_len=[0-9]{5}, "
        r"max_reply_len=[0-9]{5}",
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstances', count=3, "
        r"exception_count=0, avg_time=.+min_time=.+max_time=.+"
        r"avg_server_time=.+min_server_time.+max_server_time=.+"
        r"max_reply_len=[0-9]{5}",
        stat_repr)

    assert re.search(
        r"OperationStatistic\(name='EnumerateInstanceNames', count=1, "
        r"exception_count=0, avg_time=[.0-9]+, min_time=[.0-9]+, "
        r"max_time=[.0-9]+, avg_server_time=[.0-9]+, min_server_time="
        r"[.0-9]+.+max_server_time=[.0-9]+, avg_request_len=[.0-9]+"
        r".+max_reply_len=[0-9]{5}",
        stat_repr)

    # test formatted output

    report = statistics.formatted()

    assert re.search(
        r'Count ExcCnt +Time \[s\] +ServerTime \[s\] +RequestLen \[B\] '
        r'+ReplyLen \[B\] +Operation',
        report)

    assert re.search(
        r' +Avg +Min +Max +Avg +Min +Max +Avg +Min +Max',
        report)

    assert re.search(
        r"3     0 +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +"
        r"[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]{5} "
        r"EnumerateInstances",
        report)

    assert re.search(
        r"1     0 +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +[.0-9]+ +"
        r"[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]{5} "
        r"EnumerateInstanceNames",
        report)
